/*
This program is confidential and proprietary to MRJ Technology Solutions
and may not be reproduced, published or disclosed to others without
written authorization from MRJ.

Copyright (c) 1998 MRJ Technology Solutions, Inc. All Rights Reserved.

This program is derived from prior work some of which was originally
written as a joint project between the Numerical Aerospace Simulation
(NAS) Systems Division of NASA Ames Research Center and the National
Energy Research Supercomputer Center (NERSC) of Lawrence Livermore
National Laboratory.
*/

#include <pbs_config.h>   /* the master config generated by configure */

#include 	<stdio.h>
#include 	<stdlib.h>
#include	<unistd.h>
#include 	<string.h>
#include 	<ctype.h>
#include 	<errno.h>
#include	<fcntl.h>
#include	<netdb.h>
#include 	<sys/types.h>
#include	<netinet/in.h>

#include	"portability.h"
#include	"libpbs.h"
#include	"server_limits.h"
#include 	"svrfunc.h"
#include	"list_link.h"
#include	"attribute.h"
#include	"resource.h"
#include	"net_connect.h"
#include	"work_task.h"
#include	"job.h"
#include	"log.h"
#include	"pbs_nodes.h"
#include	"rpp.h"
#include	"dis.h"
#include	"dis_init.h"
#include	"resmon.h"

#if !defined(H_ERRNO_DECLARED)
extern int h_errno;
#endif
static char ident[] = "@(#) $RCSfile: node_manager.c,v $ $Revision: 2.0 $";

int	 	 svr_totnodes = 0;	/* total number nodes defined       */
int		 svr_clnodes  = 0;	/* number of cluster nodes	    */
int		 svr_tsnodes  = 0;	/* number of time shared nodes	    */
int		 svr_chngNodesfile = 0;	/* 1 signals want nodes file update */
					/* on server shutdown, (qmgr mods)  */
struct pbsnode **pbsnlist = NULL;
static int	 svr_numnodes = 0;	/* number nodes currently available */
static int	 exclusive;		/* node allocation type */

static char      extra_parm[] = "extra parameter(s)";
static char      no_parm[]    = "required parameter not found";
static FILE 	*nstatef = NULL;

extern int	 server_init_type;
extern int	 has_nodes;

extern int	ctnodes A_((char *));
extern resource_t next_resource_tag;
extern char	*path_home;
extern char	*path_nodes;
extern char	*path_nodes_new;
extern char	*path_nodestate;

#define		SKIP_NONE	0
#define		SKIP_EXCLUSIVE	1
#define		SKIP_ANYINUSE	2

/*
 * Tree search generalized from Knuth (6.2.2) Algorithm T just like
 * the AT&T man page says.
 *
 * The tree_t structure is for internal use only, lint doesn't grok it.
 *
 * Written by reading the System V Interface Definition, not the code.
 *
 * Totally public domain.
 */
/*LINTLIBRARY*/

/*
**      Modified by Tom Proett <proett@nas.nasa.gov> for PBS.
*/

typedef struct tree_t {
	u_long		key;
	struct pbsnode	*nodep;
	struct tree_t	*left, *right;
} tree;

tree	*ipaddrs = NULL;	/* tree of ip addrs */
tree	*streams = NULL;	/* tree of stream numbers */

/* find value in tree, return NULL if not found */
struct	pbsnode	*
tfind(key, rootp)
	const u_long	key;		/* key to be located */
	tree		**rootp;	/* address of tree root */
{
	register tree	*q;

	if (rootp == (struct tree_t **)0)
		return NULL;
	while (*rootp != NULL) {		/* Knuth's T1: */
		if (key == (*rootp)->key)	/* T2: */
			return (*rootp)->nodep;	/* we found it! */
		rootp = (key < (*rootp)->key) ?
			&(*rootp)->left :	/* T3: follow left branch */
			&(*rootp)->right;	/* T4: follow right branch */
	}
	return NULL;
}

void
tinsert(key, nodep, rootp)
	const	u_long	key;		/* key to be located */
	struct	pbsnode	*nodep;
	tree		**rootp;	/* address of tree root */
{
	register tree	*q;

	DBPRT(("tinsert: %lu %s stream %d\n", key,
		nodep->last->name, nodep->stream))

	if (rootp == (struct tree_t **)0)
		return;
	while (*rootp != (struct tree_t *)0) {	/* Knuth's T1: */
		if (key == (*rootp)->key)	/* T2: */
		    return;			/* we found it! */
		rootp = (key < (*rootp)->key) ?
			&(*rootp)->left :	/* T3: follow left branch */
			&(*rootp)->right;	/* T4: follow right branch */
	}
	q = (tree *) malloc(sizeof(tree));	/* T5: key not found */
	if (q != NULL) {			/* make new node */
		*rootp = q;			/* link new node to old */
		q->key = key;			/* initialize new node */
		q->nodep = nodep;
		q->left = q->right = NULL;
	}
	return;
}

/* delete node with given key */
void *
tdelete(key, rootp)
	const u_long	key;		/* key to be deleted */
	tree		**rootp;	/* address of the root of tree */
{
	tree		*p;
	register tree	*q;
	register tree	*r;

	DBPRT(("tdelete: %lu\n", key))
	if (rootp == (struct tree_t **)0 || (p = *rootp) == (struct tree_t *)0)
		return NULL;
	while (key != (*rootp)->key) {
		p = *rootp;
		rootp = (key < (*rootp)->key) ?
			&(*rootp)->left :		/* left branch */
			&(*rootp)->right;		/* right branch */
		if (*rootp == NULL)
			return ((void *)0);		/* key not found */
	}
	r = (*rootp)->right;				/* D1: */
	if ((q = (*rootp)->left) == NULL)		/* Left */
		q = r;
	else if (r != (struct tree_t *)0) {		/* Right is null? */
		if (r->left == (struct tree_t *)0) {	/* D2: Find successor */
			r->left = q;
			q = r;
		}
		else {		/* D3: Find (struct tree_t *)0 link */
			for (q = r->left; q->left != NULL; q = r->left)
				r = q;
			r->left = q->right;
			q->left = (*rootp)->left;
			q->right = (*rootp)->right;
		}
	}
	free((struct tree_t *) *rootp);     /* D4: Free node */
	*rootp = q;                         /* link parent to new node */
	return(p);
}

void
tfree(rootp)
	tree            **rootp;
{
	if (rootp == NULL || *rootp == NULL)
		return;
	tfree(&(*rootp)->left);
	tfree(&(*rootp)->right);
	free(*rootp);
	*rootp = NULL;
}

/*
**	Start a standard inter-server message.
*/
int
is_compose(stream, command)
     int	stream;
     int	command;
{
	int	ret;

	if (stream < 0)
		return DIS_EOF;
	DIS_rpp_reset();

	ret = diswsi(stream, IS_PROTOCOL);
	if (ret != DIS_SUCCESS)
		goto done;
	ret = diswsi(stream, IS_PROTOCOL_VER);
	if (ret != DIS_SUCCESS)
		goto done;
	ret = diswsi(stream, command);
	if (ret != DIS_SUCCESS)
		goto done;
	return DIS_SUCCESS;

 done:
	DBPRT(("is_compose: send error %s\n", dis_emsg[ret]))
	return ret;
}

/*
**	We got an EOF on a stream.
*/
void
stream_eof(stream, ret)
    int		stream;
    int		ret;
{
	static	char	id[] = "stream_eof";
	struct	pbsnode		*np;
	extern	int		pbs_rm_port;

	rpp_close(stream);

	/* find who the stream belongs to and mark down */
	if ((np = tfind((u_long)stream, &streams)) != NULL) {
		DBPRT(("%s: %s down\n", id, np->last->name))
		np->inuse |= INUSE_DOWN;
		np->stream = -1;
		tdelete((u_long)stream, &streams);
	}
	return;
}

/*
**	Send a ping to any node that is in an unknown stat.
**	If wt_parm1 is NULL, set up a worktask to ping again.
*/
void
ping_nodes(ptask)
    struct	work_task	*ptask;
{
	static	char	id[] = "ping_nodes";
	struct	pbsnode		*np;
	struct	sockaddr_in	*addr;
	int			i, ret, com;
	extern	time_t		time_now;
	extern	int		pbs_rm_port;

	DBPRT(("%s: entered\n", id))
	for (i=0; i<svr_totnodes; i++) {
		np = pbsnlist[i];

		if (np->inuse & (INUSE_DELETED|INUSE_OFFLINE))
		    continue;

		if (np->inuse & (INUSE_JOB|INUSE_JOBSHARE)) {
		   if (!(np->inuse & INUSE_NEEDS_HELLO_PING))
			continue;
		}

		if (np->stream < 0) {
			np->stream = rpp_open(np->last->name, pbs_rm_port);
			np->inuse |= INUSE_DOWN;
			if (np->stream == -1) {
				sprintf(log_buffer, "rpp_open to %s",
					np->last->name);
				log_err(errno, id, log_buffer);
				continue;
			}
			com = IS_HELLO;
			tinsert((u_long)np->stream, np, &streams);
		}
		else
			com = IS_NULL;

		DBPRT(("%s: ping %s\n", id, np->last->name))
		if (np->inuse & INUSE_NEEDS_HELLO_PING)
		    com = IS_HELLO;

		ret = is_compose(np->stream, com);
		if (ret == DIS_SUCCESS) {
			if (rpp_flush(np->stream) == 0)
			    continue;
			ret = DIS_NOCOMMIT;
		}

		addr = rpp_getaddr(np->stream);
		sprintf(log_buffer, "%s %d to %s(%s)",
			dis_emsg[ret], errno, np->last->name,
			(addr == NULL) ? "unknown" : netaddr(addr));
		log_err(-1, id, log_buffer);
		rpp_close(np->stream);
		tdelete((u_long)np->stream, &streams);
		np->stream = -1;
		np->inuse |= INUSE_DOWN;
	}

	if (ptask->wt_parm1 == NULL) {
		if (server_init_type == RECOV_HOT)
			i = 12;		/* rapid ping rate while hot restart */
		else
			i = 600;	/* relaxed ping rate for normal run  */
		(void)set_task(WORK_Timed, time_now+i, ping_nodes, NULL);
	}
}

/*
**	Input is coming from another server (MOM) over a DIS rpp stream.
**	Read the stream to get a Inter-Server request.
*/
void
is_request(stream, version)
     int	stream;
     int	version;
{
	static	char		id[] = "is_request";
	int			command = 0;
	int			ret = DIS_SUCCESS;
	int			i, j;
	unsigned long		ipaddr;
	struct	sockaddr_in	*addr;
	struct	pbsnode		*np, *node;
	extern	time_t		time_now;

	DBPRT(("%s: stream %d version %d\n", id, stream, version))
	addr = rpp_getaddr(stream);
	if (version != IS_PROTOCOL_VER) {
		sprintf(log_buffer, "protocol version %d unknown from %s",
			version, netaddr(addr));
		log_err(-1, id, log_buffer);
		rpp_close(stream);
		return;
	}

	/* check that machine is known */
	DBPRT(("%s: connect from %s\n", id, netaddr(addr)))
	if ((node = tfind((u_long)stream, &streams)) != NULL)
		goto found;

	ipaddr = ntohl(addr->sin_addr.s_addr);
	if ((node = tfind(ipaddr, &ipaddrs)) != NULL) {
		if (node->stream >= 0) {
			DBPRT(("%s: stream %d from %s already open on %d\n",
				id, stream, node->last->name, node->stream))
			rpp_close(stream);
			rpp_close(node->stream);
			tdelete((u_long)node->stream, &streams);
			node->inuse = INUSE_UNKNOWN;
			node->stream = -1;
			/* do a ping in 5 seconds */
			(void)set_task(WORK_Timed, time_now+5,
				ping_nodes, node);
			return;
		}
		node->stream = stream;
		tinsert((u_long)stream, node, &streams);
		goto found;
	}
	sprintf(log_buffer, "bad attempt to connect from %s", netaddr(addr));
	log_err(-1, id, log_buffer);
	rpp_close(stream);
	return;

 found:
	command = disrsi(stream, &ret);
	if (ret != DIS_SUCCESS)
		goto err;

	switch (command) {

	case IS_NULL:		/* a ping from server */
		DBPRT(("%s: IS_NULL\n", id))
		break;

	case IS_HELLO:
		DBPRT(("%s: IS_HELLO\n", id))
		ret = is_compose(stream, IS_CLUSTER_ADDRS);
		if (ret != DIS_SUCCESS)
			goto err;
		for (i=0; i<svr_totnodes; i++) {
			np = pbsnlist[i];

			if (np->inuse & INUSE_DELETED)
			    continue;

			DBPRT(("%s: %s(%d)\n", id, np->last->name, i))

			for (j=0; np->addrs[j]; j++) {
				u_long	ipaddr = np->addrs[j];

				DBPRT(("%s: %d\t%ld.%ld.%ld.%ld\n", id, j,
					(ipaddr & 0xff000000) >> 24,
					(ipaddr & 0x00ff0000) >> 16,
					(ipaddr & 0x0000ff00) >> 8,
					(ipaddr & 0x000000ff) ))
				ret = diswul(stream, ipaddr);
				if (ret != DIS_SUCCESS)
					goto err;
			}
		}
		rpp_flush(stream);
		break;

	default:
		sprintf(log_buffer, "unknown command %d sent from %s",
			command, node->last->name);
		log_err(-1, id, log_buffer);
		goto err;
	}

	node->inuse &= ~(INUSE_DOWN|INUSE_UNKNOWN|INUSE_NEEDS_HELLO_PING);
	rpp_eom(stream);
	return;

 err:
	/*
	** We come here if we got a DIS write error.
	*/
	sprintf(log_buffer, "%s from %s(%s)", dis_emsg[ret],
		node->last->name, netaddr(addr));
	log_err(-1, id, log_buffer);
	rpp_close(stream);
	node->inuse |= INUSE_DOWN;

	return;
}

void write_node_state()
{
	struct pbsnode *np;
	static char *fmt = "%s %d\n";
	int	i;

	DBPRT(("write_node_state: entered\n"))
	if (nstatef != NULL) {
		(void)fseek(nstatef, 0L, SEEK_SET);	/* rewind and clear */
		(void)ftruncate(fileno(nstatef), (off_t)0);
	} else {
		/* need to open for first time, temporary-move to pbsd_init */
	
		if ((nstatef = fopen(path_nodestate, "w+")) == NULL) {
			log_err(errno, "write_node_state", "could open file");
			return;
		}
	}
	/*
	**	The only state that carries forward is if the
	**	node has been marked offline.
	*/
	for (i=0; i<svr_totnodes; i++) {
		np = pbsnlist[i];

		if (np->inuse & INUSE_DELETED)
			continue;
		if (np->inuse & INUSE_OFFLINE)
			fprintf(nstatef, fmt, np->last->name, np->inuse);
	}
	(void)fflush(nstatef);
	
}
	
/*
 * free_prop - free list of prop structures created by proplist()
 */

static void free_prop(prop)
	struct prop *prop;
{
	struct prop *pp;

	for (pp = prop; pp; pp = prop) {
		prop = pp->next;
		free(pp->name);
		free(pp);
	}
}
	

/*
 * unreserve - unreserve nodes
 *
 *	If handle is set to a existing resource_t, then release all nodes
 *	associated with that handle, otherwise, (this is dangerous) 
 *	if handle == RESOURCE_T_ALL, release all nodes period.
 */

void
node_unreserve(handle)
	resource_t handle;
{
	struct  pbsnode *np;
	int		i;

	/* clear old reserve */
	for (i=0; i<svr_totnodes; i++) {
		np = pbsnlist[i];

		if (np->inuse & INUSE_RESERVE) {
			if ((handle==RESOURCE_T_ALL) || (handle==np->allocto))
				np->inuse &= ~INUSE_RESERVE;
		}
	}
}

/*
**	Look through the property list and make sure that all
**	those marked are contained in the node.
*/
static
int
hasprop(node, props)
    struct	pbsnode	*node;
    struct	prop	*props;
{
	struct  prop    *need;

	for (need=props; need; need=need->next) {
		struct	prop	*pp;

		if (need->mark == 0)	/* not marked, skip */
			continue;

		for (pp=node->first; pp; pp=pp->next) {
			if (strcmp(pp->name, need->name) == 0)
				break;		/* found it */
		}
		if (pp == NULL)
			return 0;
	}
	return 1;
}

/*
**	Mark the properties of a node that match the marked
**	properties given.
*/
static
void
mark(node, props)
    struct	pbsnode	*node;
    struct	prop	*props;
{
	struct  prop    *set, *pp;

	for (pp=node->first; pp; pp=pp->next) {
		pp->mark = 0;
		for (set=props; set; set=set->next) {
			if (set->mark == 0)
				continue;
			if (strcmp(pp->name, set->name) == 0) {
				pp->mark = 1;
				break;
			}
		}
	}
}

#define	RECURSIVE_LIMIT	3
/*
**	Search for a node which contains glorf.
**	skip indicates which nodes to pass over for this search.
**	Don't do any recursive calls deeper than RECURSIVE_LIMIT.
*/
static
int
search(glorf, skip, order, depth)
    struct	prop	*glorf;
    int			skip;
    int			order;
    int			depth;
{
	static	int	pass = INUSE_OFFLINE|INUSE_DOWN|
				INUSE_RESERVE|INUSE_UNKNOWN|INUSE_DELETED;
	struct	pbsnode	*np;
	int	found;
	int	i;

	if (++depth == RECURSIVE_LIMIT)
		return 0;

	/* look for nodes we haven't picked already */
	for (i=0; i<svr_totnodes; i++) {
		np = pbsnlist[i];

		if (np->ntype != NTYPE_CLUSTER)	/* ignore time share nodes */
			continue;
		if (np->flag != okay)
			continue;
		if ((skip == SKIP_ANYINUSE) && 
		    np->inuse)			/* ignore any nodes in use */
			continue;
		if ((skip == SKIP_EXCLUSIVE) && 
		    (np->inuse & ~INUSE_JOBSHARE))  /* ignore nodes in use */
			continue;		    /* not shared	   */
		if (!hasprop(np, glorf))
			continue;

		np->flag = thinking;
		mark(np, glorf);
		np->order = order;
		return 1;
	}

	if (glorf == NULL)			/* no property */
		return 0;			/* can't retry */

	/* try re-shuffling the nodes to get what we want */
	for (i=0; i<svr_totnodes; i++) {
		np = pbsnlist[i];

		if (np->ntype != NTYPE_CLUSTER)	/* ignore time share nodes */
			continue;
		if (np->flag != thinking)
			continue;
		if (skip == SKIP_EXCLUSIVE &&
				(np->inuse & (INUSE_JOB|pass)))
			continue;
		if (skip == SKIP_ANYINUSE &&
				(np->inuse & (INUSE_JOB|INUSE_JOBSHARE|pass)))
			continue;
		if (!hasprop(np, glorf))
			continue;

		np->flag = conflict;
		found = search(np->first, skip, np->order, depth);
		np->flag = thinking;
		if (found) {
			mark(np, glorf);
			np->order = order;
			return 1;
		}
	}
	return 0;	/* not found */
}

/*
**	Parse a number in a spec.
**	Return 0 if okay, 1 if no number exists, -1 on error
*/
static
int
number(ptr, num)
    char	**ptr;
    int		*num;
{
	char	holder[80];
	int	i = 0;
	char	*str = *ptr;

	while (isdigit(*str))
		holder[i++] = *str++;

	if (i == 0)
		return 1;

	holder[i] = '\0';
	if ((i = atoi(holder)) == 0) {
		sprintf(log_buffer, "zero illegal");
		return -1;
	}

	*ptr = str;
	*num = i;
	return 0;
}

/*
**	Check string to see if it is a legal property name.
**	If not, return 1.
*/
static
int
property(ptr, prop)
    char	**ptr;
    char	**prop;
{
	static	char	name[80];
	char*	str = *ptr;
	int	i = 0;

	if (!isalpha(*str)) {
		sprintf(log_buffer,
			"first character of property (%s) not a letter", str);
		return 1;
	}
	while (isalnum(*str) || *str == '-' || *str == '.')
		name[i++] = *str++;

	name[i] = '\0';
	*prop = (i == 0) ? NULL : name;
	*ptr = str;
	return 0;
}

/*
**	Create a property list from a string.
**	Return 0 if all is well, 1 otherwise.
*/
static
int
proplist(str, plist)
    char		**str;
    struct	prop	**plist;
{
	struct	prop	*pp;
	char		*pname;

	for (;;) {
		if (property(str, &pname))
			return 1;
		if (pname == NULL)
			break;
		pp = (struct prop *)malloc(sizeof(struct prop));
		pp->mark = 1;
		pp->name = strdup(pname);
		pp->next = *plist;
		*plist = pp;
		if  (**str != ':')
			break;
		(*str)++;
	}
	return 0;
}


/*
 *	Evaluate one element in a node spec.
 *
 *	Return 1 if it can be satisfied
 *	       0 if it cannot be completly satisfied. (not used now)
 *	      -1 if error - can never be satisfied.
 */
static
int
listelem(str, order)
    char	**str;
    int		order;
{
	int	num = 1;
	int	i, hit;
	int	ret = -1;
	struct	prop	*pp, *prop = NULL;
	struct	pbsnode	*np;

	if ((i = number(str, &num)) == -1)	/* get number */
		return ret;

	if (i == 0) {				/* number exists */
		if (**str == ':') {		/* there are properties */
			(*str)++;
			if (proplist(str, &prop))
				return ret;
		}
	}
	else {					/* no number */
		if (proplist(str, &prop))	/* must be a prop list */
			return ret;		/* with no number in front */
	}

	/* count number of nodes with the requested property */
	hit = 0;
	for (i=0; i<svr_totnodes; i++) {
		np = pbsnlist[i];

		if (hasprop(np, prop))
			hit++;
		if (hit == num)
			break;		/* found enough */
	}

	if (hit < num)			/* can never be satisfied */
		goto done;

	/*
	** Find an initial set of nodes to satisfy the request.
	** Go ahead and use any nodes no mater what state they are in.
	*/
	for (i=0; i<num; i++) {
		if (search(prop, SKIP_NONE, order, 0))
			continue;

		goto done;		/* can never be satisfied */
	}
	ret = 1;

done:
	free_prop(prop);
	return ret;
}

/*
**	Add the "global" spec to every sub-spec in "spec".
*/
static
char    *
mod_spec(spec, global)
    char	*spec;
    char	*global;
{
	static	char	line[512];
	char	*cp;
	int	len;

	len = strlen(global);
	cp = line;
	while (*spec) {
		if (*spec == '+') {
			*cp++ = ':';
			strcpy(cp, global);
			cp += len;
		}
		*cp++ = *spec++;
	}
	*cp++ = ':';
	strcpy(cp, global);

	return line;
}

/* cntjons - count jobs on (shared) nodes */

static int
cntjons(pn)
    struct pbsnode *pn;
{
	int ct = 0;
	struct jobinfo *pj = pn->jobs;

	while (pj) {
		++ct;
		pj = pj->next;
	}
	return (ct);
}

/*
 * nodecmp - compare two nodes for sorting 
 *	For "exclusive", put free nodes first.
 *	For "shared", put current shared with fewest jobs first,
 *		then free nodes, and others last
 */

static int
nodecmp(aa, bb)
    const	void	*aa;
    const	void	*bb;
{
	struct	pbsnode	*a = *(struct pbsnode **)aa;
	struct	pbsnode	*b = *(struct pbsnode **)bb;
	int	aprim, bprim;

	if (exclusive) {	/* best is free */
		aprim = (a->inuse == INUSE_FREE) ? 1 : 2;
		bprim = (b->inuse == INUSE_FREE) ? 1 : 2;
	}
	else {			/* best is shared with fewest jobs */
		aprim = (a->inuse == INUSE_JOBSHARE) ? cntjons(a) :
				((a->inuse == INUSE_FREE) ? 21 : 1000);
		bprim = (b->inuse == INUSE_JOBSHARE) ? cntjons(b) :
				((b->inuse == INUSE_FREE) ? 21 : 1000);
	}
	if (aprim == bprim)
		return (a->nprops - b->nprops);
	else
		return (aprim - bprim);
}

/*
 *	Test a node specification.  
 *
 *	Return >0 - number of nodes counted in the spec if it works,
 *	        0 - if it cannot be satisfied,
 *	       -1 - if it can never be satisfied.
 *	Okay to bail early if "early" is true.
 */
static int
node_spec(spec, early)
    char	*spec;
    int		 early;
{
	static	char	id[] = "node_spec";
	struct	pbsnode	*np;
	char	*str, *globs, *cp, *hold;
	int	i, num;
	int	rv;
	static char shared[] = "shared";

	DBPRT(("%s: entered\n", id))
	exclusive = 1;	/* by default all nodes are requested exclusively */
	spec = strdup(spec);
	if ((globs = strchr(spec, '#')) != NULL) {
		*globs++ = '\0';
		globs = strdup(globs);
		while ((cp = strrchr(globs, '#')) != NULL) {
			*cp++ = '\0';
			if (strcmp(cp, shared) != 0) {
				hold = strdup(mod_spec(spec, cp));
				free(spec);
				spec = hold;
			} else {
				exclusive = 0;
			}
		}
		if (strcmp(globs, shared) != 0) {
			hold = strdup(mod_spec(spec, globs));
			free(spec);
			spec = hold;
		} else {
			exclusive = 0;
		}
		free(globs);
	}
	str = spec;
		
	num = ctnodes(str);
	if (num > svr_clnodes) {
		free(spec);
		return -1;
	}

	/* sort nodes by state and number of attributes */
	qsort(pbsnlist, svr_totnodes, sizeof(struct pbsnode *), nodecmp);

	/* reset nodes to okay */
	svr_numnodes = 0;
	for (i=0; i<svr_totnodes; i++) {
		np = pbsnlist[i];
		if (np->inuse & INUSE_DELETED)
		    continue;

		np->flag = okay;
		DBPRT(("%s: %s inuse 0x%x nprops %d\n", id,
			np->last->name, np->inuse, np->nprops))
		if (np->inuse & (INUSE_OFFLINE|
				 INUSE_DOWN|
				 INUSE_RESERVE|
				 INUSE_JOB|
				 INUSE_UNKNOWN))
			continue;
		if (np->ntype == NTYPE_CLUSTER)
			svr_numnodes++;
	}

	for (i=1;; i++) {
		if ((rv = listelem(&str, i)) <= 0) {
			free(spec);
			return rv;
		}
		if (*str != '+')
			break;
		str++;
	}
	i = (int)*str;
	free(spec);

	if (i)					/* garbled list */
		return -1;
	if (num > svr_numnodes && early)	/* temp fail */
		return 0;
/*
 * 	At this point we know the spec is legal.
 *	Here we find a replacement for any nodes chosen above
 *	that were already inuse.
 */
	for (i=0; i<svr_totnodes; i++) {
		np = pbsnlist[i];

		if (np->ntype != NTYPE_CLUSTER)
			continue;	
		if (np->flag != thinking)
			continue;
		if (np->inuse == INUSE_FREE)	/* this one is okay */
			continue;
		if ((np->inuse & INUSE_JOBSHARE) && !exclusive)
			continue;		/* shared node */

		if ( search(np->first,
				exclusive ? SKIP_ANYINUSE : SKIP_EXCLUSIVE,
				np->order, 0)) {
			np->flag = okay;
			continue;
		}
		if (early)
			return 0;
		num = 0;
	}

	return num;	/* spec ok */
}

/*
**	Read the file, "nodes", containing the list of properties for each node.
**	The list of nodes is formed with "pbsnlist" as the head.
**	Return -1 on error, 0 otherwise.
**
**	Read the node state file, "node_state", for any "down" or "offline"
**	conditions which should be set in the nodes.  This is likely to
**	change to only dealing with "offline" as "down" will be detected.
**
**	If routine returns -1, then "log_buffer" contains a message to
**	be logged.
*/

int
setup_nodes()
{
	static	char	id[] = "setup_nodes";
	FILE	*nin;
	char	line[256];
	char	*pcolon;
	char	*token;
	int	i, num, linenum, numprops, narray;
	short	ntype;
	struct	hostent	*hp, *gethostbyname();
	struct	in_addr	addr;
	struct	pbsnode	*np;
	struct	prop	*pp, *first, *last;
	extern	char	*path_home;

	DBPRT(("%s: entered\n", id))
	if ((nin = fopen(path_nodes, "r")) == NULL) {
		log_event(PBSEVENT_ADMIN, PBS_EVENTCLASS_SERVER, "nodes",
			  "No Node description file");
		return(0);
	}

	next_resource_tag = time(0);	/* initialize next resource handle */

	if (pbsnlist == NULL) {
		narray = 16;
		pbsnlist = (struct pbsnode **)
			malloc(sizeof(struct pbsnode *) * 16);
	}
	else {
		/* free each element TODO */
		narray = svr_totnodes;
	}
	tfree(&streams);
	tfree(&ipaddrs);

	svr_totnodes = 0;
	for (linenum=1; fgets(line, sizeof(line), nin); linenum++) {
		if (line[0] == '#')	/* comment */
			continue;

		first = last = NULL;
		numprops = 0;
		for (i=0; i<sizeof(line);) {
			pcolon = 0;
			if (line[i] == '\n' || line[i] == '\0')
				break;
			if (!isalpha(line[i])) {
				sprintf(log_buffer,
				"token doesn't start with alpha on line %d",
					linenum);
				fclose(nin);
				return(-1);
			}
			/*
			** remember the start place for the token
			** and skip to the first whitespace
			*/
			token = &line[i];
			for (; i<sizeof(line); i++) {
				char	c = line[i];

				/* test all the legal chars */
				if (isalnum(c))
					continue;
				if (c == '-')
					continue;
				if (c == '.')
					continue;
				if (( last == NULL) && (c == ':')) {
					pcolon = &line[i];
					continue;
				}

				/* stop after getting to a space char */
				if (isspace(c))
					break;

				sprintf(log_buffer,
					"node file bad char '%c' in line %d",
					c, linenum);
				fclose(nin);
				return(-1);
			}
			if (i == sizeof(line)) {
				sprintf(log_buffer,
					"line overflow on %d", linenum);
				fclose(nin);
				return(-1);
			}

			/* terminate the token with a '\0' */
			
			line[i++] = '\0';
			if (pcolon != NULL)
				*pcolon   = '\0';

			/*
			** put token into prop list and
			** skip over any space to the start of the
			** next token
			*/
			pp = (struct prop *)malloc(sizeof(struct prop));
			pp->next = first;
			first = pp;
			if (last == NULL) {
				last = pp;
				if ((pcolon != NULL) &&
				    (strcmp(pcolon+1,TIMESHARED_SUFFIX) == 0)) {
					ntype = NTYPE_TIMESHARED;
					svr_tsnodes++;
				} else {
					ntype = NTYPE_CLUSTER;
					svr_clnodes++;
				}
			}
			pp->mark = 0;
			pp->name = strdup(token);
			numprops++;

			for (; i<sizeof(line); i++) {
				if (!isspace(line[i]))
					break;
			}
		}

		/* alloc space in array */
		if (svr_totnodes == narray) {
			narray *= 2;
			pbsnlist = (struct pbsnode **)realloc(pbsnlist,
					sizeof(struct pbsnode *) * narray);
		}
		np = (struct pbsnode *)malloc(sizeof(struct pbsnode));
		pbsnlist[svr_totnodes++] = np;
		np->jobs    = NULL;
		np->flag    = okay;
		np->nprops  = numprops;
		np->stream  = -1;
		np->inuse = INUSE_UNKNOWN;
		np->ntype = ntype;
		np->order   = 0;
		np->allocto = 0;
		np->first   = first;
		np->last    = last;

		DBPRT(("%s: %d: %s\n", id, svr_totnodes, np->last->name))

		if ((hp = gethostbyname(np->last->name)) == NULL) {
			sprintf(log_buffer, "host %s not found",
					np->last->name);
			return -1;
		}
		memcpy((char *)&addr, hp->h_addr, hp->h_length);
		if (hp->h_addr_list[1] == NULL) {
			char	*hname;

			if ((hp = gethostbyaddr((void *)&addr,
					sizeof(struct in_addr),
					hp->h_addrtype)) == NULL) {
				sprintf(log_buffer,
					"addr not found h_errno=%d errno=%d",
					h_errno, errno);
				return -1;
			}
			hname = (char *)strdup(hp->h_name);
			if ((hp = gethostbyname(hname)) == NULL) {
				sprintf(log_buffer,
					"bad cname %s, h_errno=%d errno=%d",
					hname, h_errno, errno);
				free(hname);
				return -1;
			}
			free(hname);
		}
		for (i=0; hp->h_addr_list[i]; i++);
		np->addrs = (u_long *)malloc(sizeof(u_long) * (i+1));
		for (i=0; hp->h_addr_list[i]; i++) {
			u_long	ipaddr;

			memcpy((char *)&addr, hp->h_addr_list[i],
					hp->h_length);
			ipaddr = ntohl(addr.s_addr);
			np->addrs[i] = ipaddr;
			tinsert(ipaddr, np, &ipaddrs);
			DBPRT(("%s: %s %lu.%lu.%lu.%lu\n", id, np->last->name,
				(ipaddr & 0xff000000) >> 24,
				(ipaddr & 0x00ff0000) >> 16,
				(ipaddr & 0x0000ff00) >> 8,
				(ipaddr & 0x000000ff) ))
		}
		np->addrs[i] = 0;
	}
	fclose(nin);
	svr_numnodes = svr_totnodes;
	pbsnlist = (struct pbsnode **)realloc(pbsnlist,
			sizeof(struct pbsnode *) * svr_numnodes);

	nin = fopen(path_nodestate, "r");
	if (nin != NULL) {
		while (fscanf(nin, "%s %d", line, &num) == 2) {
			for (i=0; i<svr_totnodes; i++) {
				np = pbsnlist[i];

				if (strcmp(np->last->name, line) == 0) {
					np->inuse = num;
					break;
				}
			}
		}	
		(void)fclose(nin);
	}

	return(svr_totnodes);
}

/*
**	Call node_spec() to allocate nodes then set them inuse.
**	Build list of allocated nodes to pass back in rtnlist.
*/
int set_nodes(pjob, spec, rtnlist)
    job		*pjob;
    char	*spec;
    char       **rtnlist;
{
	struct	howl {
		char	*name;
		int	order;
		struct	howl *	next;
	} *hp, *hlist, *curr, *prev, *nxt;
	int     i;

	static	char	*id = "set_nodes";
	struct	pbsnode	*np;
	struct	prop	*pp;
	char		*nodelist;
	resource	*pres;

	DBPRT(("%s: entered jid %s spec %s\n",
			id, pjob->ji_qs.ji_jobid, spec))
	/*
	 * allocate nodes 
	 */
	if ((i = node_spec(spec, 1)) == 0)	/* check spec */
		return PBSE_RESCUNAV;
	else if (i == -1)
		return PBSE_BADATVAL;

	if (exclusive)
		svr_numnodes -= i;
	hlist = NULL;

	for (i=0; i<svr_totnodes; i++) {
		struct	jobinfo	*jp;

		np = pbsnlist[i];
		if (np->flag != thinking)
			continue;

		/* Mark node as being IN USE */

		DBPRT(("%s: allocated node %s\n", id, np->last->name))
		np->inuse = exclusive ? INUSE_JOB : INUSE_JOBSHARE;
		jp  = (struct jobinfo *)malloc(sizeof(struct jobinfo));
		jp->next = np->jobs;
		np->jobs = jp;
		jp->job = pjob;

		/* build list of nodes ordered to match request */

		curr = (struct howl *)malloc(sizeof(struct howl));
		curr->order = np->order;
		curr->name = np->last->name;

		for (prev=NULL, hp=hlist; hp; prev=hp, hp=hp->next) {
			if (curr->order <= hp->order)
				break;
		}
		curr->next = hp;
		if (prev == NULL)
			hlist = curr;
		else
			prev->next = curr;
	}

	pjob->ji_qs.ji_svrflags |= JOB_SVFLG_HasNodes;	/* indicate has nodes */

	/* build list of allocated nodes */

	i = 1;			/* first, size list */
	for (hp=hlist; hp; hp=hp->next) {
		i += (strlen(hp->name) + 1);
	}
	nodelist = malloc(++i);
	*nodelist = '\0';
	/* now copy in name+name+... */
	for (hp=hlist; hp; hp=nxt) {
		(void)strcat(nodelist, hp->name);
		(void)strcat(nodelist, "+");
		nxt = hp->next;
		(void)free(hp);
	}
	*(nodelist + strlen(nodelist) - 1) = '\0';	/* strip trailing + */

	*rtnlist = nodelist;
	DBPRT(("Allocated to %s nodes: %s\n",
		pjob->ji_qs.ji_jobid, nodelist))
	return (PBSE_NONE);
}


/*
 * node_avail_complex - 
 *		*navail is set to number available
 *		*nalloc is set to number allocated
 *		*nresvd is set to number reserved 
 *		*ndown  is set to number down/offline
 */

int node_avail_complex(spec, navail, nalloc, nresvd, ndown)
	char	*spec;		/* In  - node spec */
	int 	*navail;	/* Out - number available */
	int	*nalloc;	/* Out - number allocated */
	int	*nresvd;	/* Out - number reserved  */
	int	*ndown;		/* Out - number down	  */
{
	int	holdnum;
	int	ret;

	holdnum = svr_numnodes;
	ret = node_spec(spec, 1);
	svr_numnodes = holdnum;

	*navail = ret;
	*nalloc = 0;
	*nresvd = 0;
	*ndown  = 0;

	return (ret);
}

/*
 * node_avail - report if nodes requested is available
 *	Does NOT even consider Time Shared Nodes 
 *
 *	Return 0 when no error in request and
 *		*navail is set to number available
 *		*nalloc is set to number allocated
 *		*nresvd is set to number reserved 
 *		*ndown  is set to number down/offline
 *	     !=0 error number when error in request
 */

int node_avail(spec, navail, nalloc, nresvd, ndown)
	char	*spec;		/* In  - node spec */
	int 	*navail;	/* Out - number available */
	int	*nalloc;	/* Out - number allocated */
	int	*nresvd;	/* Out - number reserved  */
	int	*ndown;		/* Out - number down	  */
{
	char	*id = "node_avail";
	int	ret, i;
	int	holdnum;
	struct	pbsnode	*np;
	char    *pc;
	struct prop *prop = NULL;
	register int xavail;
	register int xalloc;
	register int xresvd;
	register int xdown;


	if (spec == NULL) {
		log_event(PBSEVENT_ADMIN, PBS_EVENTCLASS_SERVER, id, "no spec");
		return (RM_ERR_NOPARAM);
	}

	pc = spec;
	if ( (strchr(spec, (int)'+') == NULL) &&
	     ((ret = number(&pc, &holdnum)) == 1) ) {

		/* A simple node spec - reply with numbers of avaiable,	*/
		/* allocated, reserved, and down nodes that match the	*/
		/* the spec, null or simple number means all		*/


		xavail = 0;
		xalloc = 0;
		xresvd = 0;
		xdown  = 0;

		/* find number of a specific type of node */

		if (*pc)
			if (proplist(&pc, &prop))
				return (RM_ERR_BADPARAM);

		for (i=0; i<svr_totnodes; i++) {
			np = pbsnlist[i];

		     if ((np->ntype == NTYPE_CLUSTER) && hasprop(np, prop)) {
			if (np->inuse == INUSE_FREE)
				++xavail;
			else if (np->inuse & (INUSE_JOB|INUSE_JOBSHARE))
				++xalloc;
			else if (np->inuse & INUSE_RESERVE)
				++xresvd;
			else if (np->inuse &
				     (INUSE_OFFLINE|INUSE_DOWN|INUSE_UNKNOWN))
				++xdown;
		    }
		}
		free_prop(prop);

		*navail = xavail;
		*nalloc = xalloc;
		*nresvd = xresvd;
		*ndown  = xdown;
		return 0;

	} else if ((ret = number(&pc, &holdnum)) == -1) {
		    /* invalid spec */
		    return (RM_ERR_BADPARAM);
	}


	/* not a simple spec - determine if supplied complex	*/
	/* node spec can be satisified from avail nodes		*/
	/* navail set to >0 if can be satified now		*/
	/*		  0 if not now but possible		*/
	/*		 -l if never possible			*/

	(void)node_avail_complex(spec, navail, nalloc, nresvd, ndown);
	return (0);
}


/*
 * node_reserve - Reserve nodes 
 *	Cannot reserver Time Shared Nodes
 *
 *	Returns: >0 - reservation suceeded, number of nodes reserved
 *		  0 - None or partial reservation
 *		 -1 - requested reservation impossible
 */

int node_reserve(nspec, tag)
	char		*nspec; /* In     - a node specification */
	resource_t	 tag;	   /* In/Out - tag for resource if reserved */
{
	static	char	id[] = "node_reserve";
	struct	pbsnode	*np;
	char		*spec;
	int		 ret_val;
	int		 i;

	DBPRT(("%s: entered\n", id))
	if ((nspec == NULL) || (*nspec == '\0')) {
		log_event(PBSEVENT_ADMIN, PBS_EVENTCLASS_SERVER, id, "no spec");
		return -1;
	}

	if (tag)
		node_unreserve(tag);    /* release so can rereserve */

	if ((ret_val = node_spec(nspec, 0)) > 0) {
		/*
		** Needed Nodes are available to be reserved.
		** Pretend they are inuse.
		*/
		for (i=0; i<svr_totnodes; i++) {
			np = pbsnlist[i];

			if (np->flag == thinking) {
				DBPRT(("hold %s\n", np->last->name))
				np->inuse |= INUSE_RESERVE;
				np->allocto = tag;
			}
		}
	} else if (ret_val == 0) {

		/* None or partial satisfaction of reservation */

		for (i=0; i<svr_totnodes; i++) {
			np = pbsnlist[i];

			if ((np->flag == thinking) &&
			    (np->inuse == INUSE_FREE))  {
				DBPRT(("reserve %s\n", np->last->name))
				np->inuse |= INUSE_RESERVE;
				np->allocto = tag;
			}
		}
	} else {

	    /* could never satisfy the reservation */

	    (void)sprintf(log_buffer, "can never reserve %s",nspec);
	    log_record(PBSEVENT_SCHED, PBS_EVENTCLASS_REQUEST, id,
			log_buffer);

	}

	return ret_val;
}

/*
 * is_ts_node - does the nodestr specify a single time share node?
 *	0 - yes
 *	1 - no, not a ts node or more than one node (name will not match)
 */

int is_ts_node(nodestr)
	char *nodestr;
{
	struct pbsnode *np;
	int		i;

	for (i=0; i<svr_totnodes; i++) {
		np = pbsnlist[i];

		if (np->ntype == NTYPE_TIMESHARED) {
			if (strcmp(nodestr, np->last->name) == 0)
				return 0;
		}
	}
	return 1;
}

/*
 * free_nodes - free nodes allocated to a job
 */

void free_nodes(pjob)
	job *pjob;
{
	static	char	id[] = "free_nodes";
	struct	pbsnode	*np;
	struct	jobinfo	*jp, *prev;
	int		i;

	DBPRT(("%s: entered\n", id))
	for (i=0; i<svr_totnodes; i++) {
		np = pbsnlist[i];

		for (prev=NULL, jp=np->jobs; jp; prev=jp, jp=jp->next) {
			if (jp->job != pjob)
				continue;

			DBPRT(("Freeing node %s from job %s\n",
				np->last->name, pjob->ji_qs.ji_jobid))
			if (prev == NULL)
				np->jobs = jp->next;
			else
				prev->next = jp->next;
			free(jp);
			if (np->jobs == NULL)
				np->inuse &= ~(INUSE_JOB|INUSE_JOBSHARE);
			break;
		}
	}
	pjob->ji_qs.ji_svrflags &= ~JOB_SVFLG_HasNodes;
}

/*
 * set_one_old - set a named node as allocated to a a job
 *
 *	Depends on the "real" node name being the last properity
 */

static void set_one_old(name, pjob, shared)
	char *name;
	job  *pjob;
	int   shared;	/* how used flag, either INUSE_JOB or INUSE_JOBSHARE */
{
	int		i;
	struct pbsnode *np;
	struct jobinfo *jp;

	for (i=0; i<svr_totnodes; i++) {
		np = pbsnlist[i];

		if (strcmp(name, np->last->name) == 0) {

		    /* Mark node as being IN USE ...
		     * Right now we don't have a way fof knowing if the
		     * node was inuse shared or exclusive - so set excl
		     * unless we find that this is the second job
		     */

		    if (np->ntype == NTYPE_CLUSTER) {

			np->inuse = shared;
			jp  = (struct jobinfo *)malloc(sizeof(struct jobinfo));
			jp->next = np->jobs;
			np->jobs = jp;
			jp->job = pjob;

		    }
		    break;
		}
	}
}
	
/*
 * set_old_nodes - set "old" nodes as in use - called from pbsd_init
 *	when recovering a job in the running state.
 */
void set_old_nodes(pjob)
	job *pjob;
{
	char *old;
	char *po;
	struct pbsnode *np;
	resource *presc;
	int   shared = INUSE_JOB;

	if ((pbsnlist != NULL) && 
	    (pjob->ji_wattr[(int)JOB_ATR_exec_host].at_flags & ATR_VFLAG_SET)) {

		/* are the nodes being used shared? Look in "neednodes" */

		presc = find_resc_entry(&pjob->ji_wattr[(int)JOB_ATR_resource],
					find_resc_def(svr_resc_def, "neednodes",
					svr_resc_size) );
		if ( (presc != NULL) && 
		     (presc->rs_value.at_flags & ATR_VFLAG_SET) ) {
			if (po = strchr(presc->rs_value.at_val.at_str, '#')) {
				if (strstr(++po, "shared") != NULL) 
					shared = INUSE_JOBSHARE;
			}
		}

		
		old = strdup(pjob->ji_wattr[(int)JOB_ATR_exec_host].at_val.at_str);
		while ((po = strrchr(old, (int)'+')) != NULL) {

			*po++ = '\0';
			set_one_old(po, pjob, shared);
		}
		set_one_old(old, pjob, shared);

		free(old);
	}
}
